from cupyd.core.etl import ETL
from cupyd.core.nodes import Extractor, Transformer, Filter, Loader, Bulker, DeBulker

__all__ = ["ETL", "Extractor", "Transformer", "Filter", "Loader", "Bulker", "DeBulker"]



from cupyd.core.graph.algorithms import topological_sort
from cupyd.core.graph.classes import Edge


def test__topological_sort(node_a, node_b, node_c, node_d, node_e, node_f, node_g, node_h, node_i):
    node_a >> node_e >> node_f >> [node_g >> node_h >> node_b, node_d]
    # no need to assign to anything but just to avoid warning on newer python versions (>= 3.12)
    _ = node_e >> node_c
    _ = node_f >> node_i

    nodes, edges = topological_sort(root_node=node_a)

    assert nodes == [
        node_a,
        node_e,
        node_f,
        node_g,
        node_h,
        node_b,
        node_d,
        node_i,
        node_c,
    ]
    assert edges == [
        Edge(node_a, node_e),
        Edge(node_e, node_f),
        Edge(node_f, node_g),
        Edge(node_g, node_h),
        Edge(node_h, node_b),
        Edge(node_f, node_d),
        Edge(node_f, node_i),
        Edge(node_e, node_c),
    ]



import pytest

from cupyd import ETL
from cupyd.core.exceptions import ETLExecutionError
from cupyd.tests.etl.nodes import ListExtractor, AdderToStr, ListLoader, CustomFilter


def test__etl_error_transformer():
    items = [0, 1, 2, "3", 4, 5]

    ext = ListExtractor(items=items)
    tf = AdderToStr()
    filter_1 = CustomFilter()
    ldr_1 = ListLoader()
    ldr_2 = ListLoader()

    ext >> tf >> [ldr_1, filter_1 >> ldr_2]
    etl = ETL(ext)

    with pytest.raises(ETLExecutionError):
        etl.run()


def test__etl_error_extractor():
    items = None

    ext = ListExtractor(items=items)  # type: ignore
    tf = AdderToStr()
    filter_1 = CustomFilter()
    ldr_1 = ListLoader()
    ldr_2 = ListLoader()

    ext >> tf >> [ldr_1, filter_1 >> ldr_2]
    etl = ETL(ext)

    with pytest.raises(ETLExecutionError):
        etl.run()



from __future__ import annotations

from abc import abstractmethod
from re import compile as re_compile
from typing import List, Union, Tuple, Iterable, Any, final, Optional

from cupyd.core.exceptions import (
    CyclicNodeError,
    NodesAlreadyConnectedError,
    NodesConnectionError,
)
from cupyd.core.models.node_exception import NodeException

__all__ = ["Node", "Edge", "SubGraph"]

NODE_NAME_REGEX = re_compile(r"[A-Z]?[a-z]+|[A-Z]{2,}(?=[A-Z][a-z]|\d|\W|$)|\d+")


class _Connectable:

    @abstractmethod
    def __rshift__(self, target: Union[_Connectable, List[_Connectable]]):
        raise NotImplementedError


class Node(_Connectable):

    def __init__(self):
        self._id: Optional[str] = None
        self._name: Optional[str] = None
        self._input: Optional[Node] = None
        self._outputs: List[Node] = []

    @final
    def __str__(self) -> str:
        if self._name:
            return self._name
        else:
            # build a snake_case representation out of the CamelCase class name
            class_name = self.__class__.__name__
            tokens = [str(token).lower() for token in NODE_NAME_REGEX.findall(class_name)]
            return "_".join(tokens)

    @final
    def __repr__(self) -> str:
        return str(self)

    @final
    def __rshift__(self, target: Union[_Connectable, Iterable[_Connectable]]) -> SubGraph:
        return _NodeConnector.connect(origin=self, target=target)

    def start(self):
        """Code to run on Node initialization."""
        pass

    def finalize(self):
        """Code to run on Node finalization."""
        pass

    def handle_exception(self, exception: NodeException):
        """Code to run on Node exception.

        By default, the finalize() method will be called, no matter the exception.
        The user may override it to add custom behaviour based on the raised Exception,
        but should remember to call finalize(), if necessary.

        Parameters
        ----------
        exception : NodeException
            Exception that occurred inside the Node.

        Returns
        -------
        None
        """

        self.finalize()

    @property
    def input(self) -> Optional[Node]:
        return self._input

    @input.setter
    def input(self, node: Node):
        if not isinstance(node, Node):
            raise TypeError(f"Unable to set input node with {node}")
        if self._input:
            raise NodesConnectionError(
                "Node input was already set. Please, for safety, create a new Node instance "
                "instead of re-using an old one"
            )
        if node == self:
            raise CyclicNodeError("Unable to connect Node with itself!")
        self._input = node

    @property
    def outputs(self) -> List[Node]:
        return self._outputs

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value: str):
        if not isinstance(value, str):
            raise TypeError
        self._name = value

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value: str):
        if not isinstance(value, str):
            raise TypeError
        self._id = value


class Edge:

    def __init__(self, origin: Node, target: Node):
        self.origin = origin
        self.target = target

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Edge):
            return (self.origin == other.origin) and (self.target == other.target)

    def __str__(self) -> str:
        return f"{self.origin} ⇒ {self.target}"

    def __repr__(self) -> str:
        return str(self)

    def to_tuple(self) -> Tuple[Node, Node]:
        return self.origin, self.target

    def to_str_tuple(self) -> Tuple[str, str]:
        return str(self.origin), str(self.target)


class SubGraph(_Connectable):
    """A subset of Nodes within a Directed Acyclic Graph (DAG)."""

    def __init__(self, root_node: Node, leaf_nodes: Iterable[Node] = None):
        if not root_node:
            raise ValueError("Cannot initialize a SubGraph without a root Node!")

        self.root_node = root_node
        self.leaf_nodes = tuple(node for node in leaf_nodes) if leaf_nodes else ()

    def __rshift__(self, target: Union[_Connectable, Iterable[_Connectable]]) -> SubGraph:
        origin = self.leaf_nodes or self.root_node

        if isinstance(origin, tuple):
            if len(origin) > 1:
                raise NodesConnectionError(
                    f"Forbidden to connect multiple Nodes {origin} into another {target}"
                )
            else:
                origin = origin[0]

        subgraph = _NodeConnector.connect(origin=origin, target=target)

        return SubGraph(root_node=self.root_node, leaf_nodes=subgraph.leaf_nodes)


class _NodeConnector:

    @classmethod
    def connect(cls, origin: Node, target: Union[_Connectable, Iterable[_Connectable]]) -> SubGraph:
        """Connect a list of Nodes to one of the following Connectable:

        a) Another Node.
        b) A SubGraph
        c) A list of Nodes and/or SubGraphs.

        Lastly, return the resulting SubGraph.
        """

        target_root_nodes, target_leaf_nodes = cls._get_nodes_from_target(target)

        for target_node in target_root_nodes:
            cls._connect_nodes(origin=origin, target=target_node)

        subgraph = SubGraph(root_node=origin, leaf_nodes=target_leaf_nodes)

        return subgraph

    @staticmethod
    def _get_nodes_from_target(
        target: Union[_Connectable, List[_Connectable]]
    ) -> Tuple[List[Node], List[Node]]:

        target_root_nodes = []
        target_leaf_nodes = []

        if isinstance(target, Iterable):
            for d in target:
                if isinstance(d, Node):
                    target_root_nodes.append(d)
                    target_leaf_nodes.append(d)
                elif isinstance(d, SubGraph):
                    target_root_nodes.append(d.root_node)
                    if d.leaf_nodes:
                        target_leaf_nodes.extend(d.leaf_nodes)
                    else:
                        target_leaf_nodes.append(d.root_node)
                else:
                    raise TypeError(f"Cannot connect to a {type(target)} target object.")
        elif isinstance(target, Node):
            target_root_nodes = [target]
            target_leaf_nodes = [target]
        elif isinstance(target, SubGraph):
            target_root_nodes = [target.root_node]
            target_leaf_nodes = target.leaf_nodes or [target.root_node]
        else:
            raise TypeError(f"Cannot connect to a {type(target)} target object.")

        return target_root_nodes, target_leaf_nodes

    @staticmethod
    def _connect_nodes(origin: Node, target: Node):
        """Connect two Nodes.

        Two checks must pass before making the connection:

            1. We are not connecting the Node with itself.
            2. Both Nodes are not already connected.
        """

        if origin == target:
            raise CyclicNodeError(f"Cannot connect a Node with itself: {origin}")

        if any(node == target for node in origin.outputs):
            raise NodesAlreadyConnectedError(f"Node {origin} already connected to {target}")

        target.input = origin
        origin.outputs.append(target)



import logging
from multiprocessing import Process
from multiprocessing import Queue as MultiprocessingQueue
from queue import Queue
from threading import Thread
from typing import List, Dict, Type

from cupyd.core.communication import (
    Connector,
    InterProcessEventFlag,
    IntraProcessConnector,
    MPCounter,
    InterruptionHandler,
)
from cupyd.core.computing.node_worker import (
    NodeWorker,
    ExtractorWorker,
    ProcessorWorker,
    DeBulkerWorker,
    BulkerWorker,
)
from cupyd.core.constants.logging import LOGGING_FORMAT
from cupyd.core.constants.sentinel_values import NO_MORE_ITEMS
from cupyd.core.graph.classes import Node
from cupyd.core.models.node_exception import NodeException
from cupyd.core.nodes import Extractor, Transformer, Loader, Filter, Bulker, DeBulker


class ETLWorker:

    def __init__(
        self,
        worker_id: str,
        segment_id: str,
        nodes: List[Node],
        counters: Dict[str, MPCounter],
        input_connector_by_node_id: Dict[str, Connector],
        output_connectors_by_node_id: Dict[str, List[Connector]],
        monitor_performance_event_by_node_id: Dict[str, InterProcessEventFlag],
        stop_event: InterProcessEventFlag,
        pause_event: InterProcessEventFlag,
        interruption_handler: InterruptionHandler,
        node_timings: MultiprocessingQueue,
        finished_workers: MultiprocessingQueue,
    ):
        super().__init__()
        self.worker_id = worker_id
        self.segment_id = segment_id
        self.nodes = nodes
        self.counters = counters
        self.input_connector_by_node_id = input_connector_by_node_id
        self.output_connectors_by_node_id = output_connectors_by_node_id
        self.monitor_performance_event_by_node_id = monitor_performance_event_by_node_id
        self.stop_event = stop_event
        self.pause_event = pause_event
        self.node_timings = node_timings
        self.interruption_handler = interruption_handler
        self.finished_workers = finished_workers

    def run(self):
        logging.basicConfig(level=logging.INFO, format=LOGGING_FORMAT, force=True)

        if isinstance(self, ETLWorkerProcess):
            self.interruption_handler.start()

        thread_by_node_id: Dict[str, NodeWorker] = {}
        finished_threads_queue = Queue()

        # start IntraProcessConnectors (InterProcessConnectors were started outside the ETLWorker)
        for connector in self.input_connector_by_node_id.values():
            if isinstance(connector, IntraProcessConnector):
                connector.start()

        # each Node will have its own thread
        for node in self.nodes:
            thread_class = self._get_node_worker_class(node=node)
            thread = thread_class(
                node=node,
                counter=self.counters.get(node.id, None),
                input_connector=self.input_connector_by_node_id.get(node.id, None),
                output_connectors=self.output_connectors_by_node_id.get(node.id, []),
                finished_threads_queue=finished_threads_queue,
                stop_event=self.stop_event,
                pause_event=self.pause_event,
                node_timings=self.node_timings,
                monitor_performance=self.monitor_performance_event_by_node_id[node.id],
            )
            thread.start()
            thread_by_node_id[node.id] = thread

        exception_by_node_id: Dict[str, NodeException] = {}

        while thread_by_node_id:
            node_id, exception = finished_threads_queue.get()

            # safely wait till the thread is completely finished
            thread_by_node_id.pop(node_id).join()

            # send sentinel value to every IntraProcessConnector that is output of the finished
            # Node, if any. InterProcessConnectors will be handled from the main process thread
            for connector in self.output_connectors_by_node_id.get(node_id, []):
                if isinstance(connector, IntraProcessConnector):
                    connector.produce(NO_MORE_ITEMS)

            if exception:
                exception_by_node_id[node_id] = exception

        # notify the main process thread this ETLWorker has finished
        self.finished_workers.put((self.worker_id, self.segment_id, exception_by_node_id))

    @staticmethod
    def _get_node_worker_class(node: Node) -> Type[NodeWorker]:
        if isinstance(node, Extractor):
            return ExtractorWorker
        elif isinstance(node, Transformer) or isinstance(node, Filter) or isinstance(node, Loader):
            return ProcessorWorker
        elif isinstance(node, Bulker):
            return BulkerWorker
        elif isinstance(node, DeBulker):
            return DeBulkerWorker
        else:
            raise TypeError(f"Invalid node type: {type(node)}")


class ETLWorkerThread(ETLWorker, Thread):
    pass


class ETLWorkerProcess(ETLWorker, Process):
    pass



import multiprocessing
import queue
from abc import abstractmethod
from copy import deepcopy
from typing import Optional, Any, List

from cupyd.core.constants.sentinel_values import NO_MORE_ITEMS


class Connector:
    """Unidirectional connection mechanism between two Nodes.

    There are 2 possible types of Connector based on where both Nodes exist. Each Connector type
    will use its own communication mechanism to share items between the two Nodes.

    These are the available Connector types:

    1) IntraProcess Connector
        * Nodes run at the same Process.
        * Communication mechanism: queue.Queue
        * Optional: can choose a maxsize for its internal queue.

    2) InterProcess Connector
        * Nodes run in different Processes from same computer.
        * Communication mechanism: multiprocessing.Queue
        * Optional: can choose a maxsize for its internal queue.

    TODO 3) Distributed Connector
        * Nodes run in different Processes from different computers.
        * Communication mechanism: TCP Socket
    """

    def __init__(self, maxsize: Optional[int] = 0):
        super().__init__()
        self._maxsize = maxsize
        self._started = False

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def produce(self, bucket: List[Any]):
        pass

    @abstractmethod
    def consume(self) -> Optional[List[Any]]:
        pass

    @abstractmethod
    def finish_producing(self, num_consumers: int):
        pass

    @abstractmethod
    def get_current_size(self) -> int:
        pass

    @abstractmethod
    def close(self):
        pass

    @property
    def started(self) -> bool:
        return self._started


class IntraProcessConnector(Connector):
    """Used for communication between Nodes at the same process pool where threading is used."""

    def __init__(self, maxsize: Optional[int] = 0):
        super().__init__(maxsize=maxsize)
        self.copy_bucket_on_produce: bool = False
        self._queue: Optional[queue.Queue] = None

    def start(self):
        self._queue = queue.Queue(maxsize=self._maxsize)
        self._started = True

    def produce(self, bucket: List[Any]):
        if self.copy_bucket_on_produce:
            bucket = deepcopy(bucket)
        self._queue.put(bucket)

    def consume(self) -> Optional[List[Any]]:
        return self._queue.get()

    def get_current_size(self) -> int:
        return self._queue.qsize()

    def finish_producing(self, num_consumers: int):
        for _ in range(num_consumers):
            self._queue.put(NO_MORE_ITEMS)

    def close(self):
        pass


class InterProcessConnector(Connector):
    """Used for communication between Nodes at different processes pools."""

    def __init__(self, maxsize: Optional[int] = 0):
        super().__init__(maxsize=maxsize)
        self._queue: Optional[multiprocessing.Queue] = None

    def start(self):
        self._queue = multiprocessing.Queue(maxsize=self._maxsize)
        self._started = True

    def produce(self, bucket: List[Any]):
        self._queue.put(bucket)

    def consume(self) -> Optional[List[Any]]:
        bucket = self._queue.get()
        return bucket

    def get_current_size(self) -> int:
        return self._queue.qsize()

    def finish_producing(self, num_consumers: int):
        for _ in range(num_consumers):
            self._queue.put(NO_MORE_ITEMS)

    def close(self):
        self._queue.close()



from traceback import TracebackException

from cupyd.core.constants.node_actions import NODE_ACTIONS


class NodeException:
    def __init__(self, exc: Exception, action: str):
        if isinstance(exc, Exception):
            self.exc = exc
        else:
            raise TypeError("Input exception isn't one!")
        if action not in NODE_ACTIONS:
            raise ValueError(f"Node action {action} not recognized")
        else:
            self.action = action

        traceback_exc = TracebackException.from_exception(exc=self.exc)
        traceback_formatted = list(traceback_exc.format())
        self.traceback_formatted = "".join(traceback_formatted)

    def __str__(self):
        return f"{self.action}: {self.exc}"

    def __repr__(self):
        return str(self)



from cupyd.core.communication.connector import (
    Connector,
    IntraProcessConnector,
    InterProcessConnector,
)
from cupyd.core.communication.counter import MPCounter
from cupyd.core.communication.event_flag import (
    EventFlag,
    IntraProcessEventFlag,
    InterProcessEventFlag,
)
from cupyd.core.communication.interruption_handler import InterruptionHandler

__all__ = [
    "Connector",
    "IntraProcessConnector",
    "InterProcessConnector",
    "EventFlag",
    "IntraProcessEventFlag",
    "InterProcessEventFlag",
    "MPCounter",
    "InterruptionHandler",
]



from abc import abstractmethod
from multiprocessing import Event as MPEvent
from threading import Event as ThreadingEvent


class EventFlag:
    """This communication mechanism will be used to transmit actions to specific ETL Nodes."""

    @abstractmethod
    def __bool__(self) -> bool:
        """Indicate if the event is True or False."""
        pass

    @abstractmethod
    def wait(self):
        """If the event is enabled, wait till its set to False."""
        pass

    @abstractmethod
    def set(self):
        """Activate the event by setting it to True."""
        pass


class InterProcessEventFlag(EventFlag):

    def __init__(self, start_set: bool = False):
        self._event = MPEvent()
        if not start_set:
            self._event.set()

    def __bool__(self) -> bool:
        return not self._event.is_set()

    def wait(self):
        self._event.wait()

    def set(self):
        if self:
            self._event.set()
        else:
            self._event.clear()


class IntraProcessEventFlag(EventFlag):

    def __init__(self):
        self._event = ThreadingEvent()
        self._event.set()

    def __bool__(self) -> bool:
        return not self._event.is_set()

    def wait(self):
        self._event.wait()

    def set(self):
        if self:
            self._event.set()
        else:
            self._event.clear()



class GraphException(Exception):
    pass


class CyclicNodeError(GraphException):
    pass


class NodesAlreadyConnectedError(GraphException):
    pass


class CyclicGraphError(GraphException):
    pass


class NodesConnectionError(GraphException):
    pass


class ETLExecutionError(Exception):
    pass


class InterruptedETL(Exception):
    pass



import typing
from collections import defaultdict
from itertools import groupby
from typing import List, Tuple

from cupyd.core.graph.classes import Node, Edge
from cupyd.core.models.etl_segment import ETLSegment
from cupyd.core.nodes import Extractor, Transformer, Loader, Filter, Bulker, DeBulker


# PUBLIC FUNCTIONS


def topological_sort(root_node: Node) -> Tuple[List[Node], List[Edge]]:
    nodes = []
    edges = []

    # if the graph contains cycles (not a DAG), it will fail when computing the topological sort.
    nodes, edges = _downstream_discovery(root_node, nodes, edges)

    return nodes, edges


def assign_names_and_ids_to_nodes(nodes: List[Node]) -> None:
    """If a Node doesn't have a name provide it one using the class name with a numbered suffix."""

    node_id = 1
    num_nodes_per_class = defaultdict(int)
    last_num_per_class = defaultdict(int)

    for node in nodes:
        node.id = f"node_{node_id}"
        node_id += 1

        if node.name is None:
            class_name = str(node)
            num_nodes_per_class[class_name] += 1

    for node in nodes:
        if node.name is None:
            class_name = str(node)
            last_num_per_class[class_name] += 1
            if num_nodes_per_class[class_name] == 1:  # single Node from that class without name
                node.name = class_name
            else:
                node.name = f"{class_name}_{last_num_per_class[class_name]}"


def get_etl_segments(nodes: List[Node], num_workers: int) -> List[ETLSegment]:
    segments = []
    segment_num = 1

    for group in _split_nodes_by_attr(nodes=nodes, attr_name="run_in_main_process"):
        for group_ in _split_nodes_if_not_consecutive(nodes=group):
            run_in_main_process = group_[0].configuration.run_in_main_process

            if run_in_main_process or isinstance(group_[0], Extractor):
                segment_num_workers = 1
            else:
                segment_num_workers = num_workers

            segments.append(
                ETLSegment(
                    id=f"segment_{segment_num}",
                    nodes=group_,
                    node_ids={node.id for node in group_},
                    num_workers=segment_num_workers,
                    run_in_main_process=run_in_main_process,
                )
            )
            segment_num += 1

    return segments


# PROTECTED FUNCTIONS


def _downstream_discovery(
    node: Node, nodes: List[Node], edges: List[Edge]
) -> Tuple[List[Node], List[Edge]]:
    """Discover unique Nodes and Edges from top to bottom, from left to right."""

    if not nodes:
        nodes = []
    if not edges:
        edges = []

    if node not in nodes:  # avoid storing duplicate Nodes when at least 2 branches are connected
        nodes.append(node)

    for descendant in node.outputs:
        edge = Edge(origin=node, target=descendant)

        if edge not in edges:
            edges.append(edge)

        nodes, edges = _downstream_discovery(descendant, nodes, edges)

    return nodes, edges


def _split_nodes_by_attr(nodes: List[Node], attr_name: str) -> List[List[Node]]:
    """This function will split Nodes based on equality of a selected Node attribute."""

    groups = [
        list(g)
        for _, g in groupby(
            sorted(nodes, key=lambda x: _get_node_attr(x, attr_name)),
            key=lambda x: _get_node_attr(x, attr_name),
        )
    ]
    return groups


def _get_ascendants(
    node: Node, ascendants: List[Node] = None, valid_nodes_ids: typing.Set[str] = None
) -> List[Node]:
    if not ascendants:
        ascendants = []

    if node.input:
        if (valid_nodes_ids and node.input.id in valid_nodes_ids) or not valid_nodes_ids:
            ascendants.append(node.input)
            ascendants.extend(_get_ascendants(node=node.input, ascendants=ascendants))

    return ascendants


def _get_descendants(
    node: Node, descendants: List[Node] = None, valid_nodes_ids: typing.Set[str] = None
) -> List[Node]:
    if not descendants:
        descendants = []

    for output_node in node.outputs:
        if (valid_nodes_ids and output_node.id in valid_nodes_ids) or not valid_nodes_ids:
            descendants.append(output_node)
            descendants.extend(_get_descendants(node=output_node, descendants=descendants))

    return descendants


def _split_nodes_if_not_consecutive(nodes: List[Node]) -> List[List[Node]]:
    node_ids_to_check = {node.id for node in nodes}
    connected_nodes_by_id = {}

    for node in nodes:
        connected_nodes_by_id[node.id] = {node.id}

        for ascendant in _get_ascendants(node=node, valid_nodes_ids=node_ids_to_check):
            if ascendant.id in node_ids_to_check:
                connected_nodes_by_id[node.id].add(ascendant.id)

        for descendant in _get_descendants(node=node, valid_nodes_ids=node_ids_to_check):
            if descendant.id in node_ids_to_check:
                connected_nodes_by_id[node.id].add(descendant.id)

    groups = []

    while connected_nodes_by_id:
        node_id, connected_nodes = connected_nodes_by_id.popitem()
        node_ids_group = {node_id}

        for node_id_, connected_nodes_ in connected_nodes_by_id.items():

            # path exists between both nodes
            if connected_nodes.intersection(connected_nodes_):
                node_ids_group.add(node_id_)

        for node_id in node_ids_group:
            connected_nodes_by_id.pop(node_id, None)

        group = []

        for node in nodes:
            if node.id in node_ids_group:
                group.append(node)

        groups.append(group)

    # extractor nodes must be alone if they don't have run_in_main_process
    for group in groups:
        num_nodes = len(group)
        if num_nodes > 1:
            for idx in range(num_nodes):
                node = group[idx]
                if isinstance(node, Extractor) and not node.configuration.run_in_main_process:
                    group.pop(idx)
                    groups.append([node])
                    return groups

    return groups


def _get_node_attr(
    node: typing.Union[Extractor, Transformer, Loader, Filter, Bulker, DeBulker],
    attr_name: str,
    default: typing.Any = None,
):
    return getattr(node.configuration, attr_name, default)



import logging
from typing import Iterator, Any, Optional

from cupyd.core.etl import ETL
from cupyd.core.nodes import Extractor, Transformer, Filter, Loader

logger = logging.getLogger("my_logger")


class IntExtractor(Extractor):

    def __init__(self, n: int):
        super().__init__()
        self.configuration.run_in_main_process = True
        self.n = n

    def start(self):
        logger.info("Extractor started!")

    def extract(self) -> Iterator[int]:
        for value in range(self.n):
            yield value


class Adder(Transformer):

    def start(self):
        pass

    def transform(self, value: int) -> int:
        if value == 500_000:
            raise ValueError("We reached the value 500.000!")
        else:
            return value + 1

    def finalize(self):
        logger.info("Adder finalized!")


class ListLoader(Loader):

    def __init__(self):
        super().__init__()
        self.configuration.run_in_main_process = True
        self.items = []

    def start(self, msg: Any = None):
        self.items = []

    def load(self, item: Any):
        self.items.append(item)


class EvenValuesFilter(Filter):

    def filter(self, item: int) -> Optional[int]:
        if item % 2 == 0:
            return None
        else:
            return item


def run_etl():
    total_items = 50_000_000

    # EXTRACTOR
    ext = IntExtractor(n=total_items)
    ext.configuration.bucket_size = 10000
    ext.configuration.run_in_main_process = False

    tf_1 = Adder()

    tf_2 = Adder()
    tf_2.configuration.run_in_main_process = True

    # FILTER NODE
    flt_1 = EvenValuesFilter()
    flt_1.configuration.run_in_main_process = True

    # LOADERS
    ldr_a = ListLoader()
    ldr_b = ListLoader()

    ext >> tf_1 >> tf_2 >> ldr_a
    tf_1 >> flt_1 >> ldr_b

    etl = ETL(extractor=ext)
    etl.run(
        workers=4,
        raise_exception=True,
        show_progress=True,
        monitor_performance=True,
        verbose=True,
    )

    logger.info(f"Total items in ldr_a {len(ldr_a.items)} | total items ldr_b {len(ldr_b.items)}")


if __name__ == "__main__":
    logging.basicConfig(format="%(asctime)s - %(message)s", level=logging.INFO)
    logging.info("Starting my script...")
    run_etl()
    logging.info("Exiting my script")



from cupyd.core.nodes.bulker import Bulker
from cupyd.core.nodes.debulker import DeBulker
from cupyd.core.nodes.extractor import Extractor
from cupyd.core.nodes.filter import Filter
from cupyd.core.nodes.loader import Loader
from cupyd.core.nodes.transformer import Transformer

__all__ = ["Extractor", "Transformer", "Filter", "Loader", "Bulker", "DeBulker"]



from pytest import fixture

from cupyd.core.graph.classes import Node


@fixture(scope="function", autouse=True)
def node_a() -> Node:
    node = Node()
    node.name = "A"
    yield node


@fixture(scope="function", autouse=True)
def node_b():
    node = Node()
    node.name = "B"
    yield node


@fixture(scope="function", autouse=True)
def node_c():
    node = Node()
    node.name = "C"
    yield node


@fixture(scope="function", autouse=True)
def node_d():
    node = Node()
    node.name = "D"
    yield node


@fixture(scope="function", autouse=True)
def node_e():
    node = Node()
    node.name = "E"
    yield node


@fixture(scope="function", autouse=True)
def node_f():
    node = Node()
    node.name = "F"
    yield node


@fixture(scope="function", autouse=True)
def node_g():
    node = Node()
    node.name = "G"
    yield node


@fixture(scope="function", autouse=True)
def node_h():
    node = Node()
    node.name = "H"
    yield node


@fixture(scope="function", autouse=True)
def node_i():
    node = Node()
    node.name = "i"
    yield node


@fixture(scope="function", autouse=True)
def node_j():
    node = Node()
    node.name = "J"
    yield node



PINK = "\033[95m"
BLUE = "\033[94m"
CYAN = "\033[96m"
GREEN = "\033[92m"
YELLOW = "\033[93m"
RED = "\033[91m"
BOLD = "\033[1m"
UNDERLINE = "\033[4m"
END = "\033[0m"

LOGGING_FORMAT = (
    "%(asctime)-28s"
    # todo: we could use the thread name to show from which node each log comes from
    # f'{BOLD}%(threadName)-20s{END}'
    "%(name)-26s"
    "%(levelname)-12s"
    "%(message)s"
)

LOGGING_MSG_PADDING = f"{'':>64}"



import pytest

from cupyd.core.exceptions import (
    NodesConnectionError,
    CyclicNodeError,
    NodesAlreadyConnectedError,
)


def test__simple_connection(node_a, node_b, node_c):
    node_a >> node_b >> node_c

    assert node_a.input is None
    assert node_b.input == node_a
    assert node_c.input == node_b

    assert node_a.outputs == [node_b]
    assert node_b.outputs == [node_c]
    assert node_c.outputs == []


def test__branched_connection(node_a, node_b, node_c):
    node_a >> [node_b, node_c]

    assert node_a.input is None
    assert node_b.input == node_a
    assert node_c.input == node_a

    assert node_a.outputs == [node_b, node_c]
    assert node_b.outputs == []
    assert node_c.outputs == []


def test__mixed_connection(node_a, node_b, node_c, node_d):
    node_a >> [node_b, node_c >> node_d]

    assert node_a.input is None
    assert node_b.input == node_a
    assert node_c.input == node_a
    assert node_d.input == node_c

    assert node_a.outputs == [node_b, node_c]
    assert node_b.outputs == []
    assert node_c.outputs == [node_d]
    assert node_d.outputs == []


def test__subgraph_left_connection(node_a, node_b, node_c, node_d):
    subgraph_1 = node_a >> [node_b, node_c]

    assert subgraph_1.root_node == node_a
    assert subgraph_1.leaf_nodes == (node_b, node_c)

    assert node_a.input is None
    assert node_b.input == node_a
    assert node_c.input == node_a
    assert node_d.input is None

    assert node_a.outputs == [node_b, node_c]
    assert node_b.outputs == []
    assert node_c.outputs == []
    assert node_d.outputs == []

    subgraph_2 = node_d >> subgraph_1

    assert subgraph_2.root_node == node_d
    assert subgraph_2.leaf_nodes == subgraph_1.leaf_nodes == (node_b, node_c)

    assert node_a.input == node_d
    assert node_d.input is None
    assert node_d.outputs == [node_a]

    assert node_a.outputs == [node_b, node_c]
    assert node_b.outputs == []
    assert node_c.outputs == []


def test__subgraph_right_connection(node_a, node_b, node_c, node_d):
    subgraph_1 = node_a >> node_b
    subgraph_2 = subgraph_1 >> node_c >> node_d

    assert subgraph_1.root_node == node_a
    assert subgraph_1.leaf_nodes == (node_b,)

    assert subgraph_2.root_node == node_a
    assert subgraph_2.leaf_nodes == (node_d,)


def test__nodes_connection_error(node_a, node_b, node_c, node_d):
    subgraph_1 = node_a >> [node_b, node_c]
    with pytest.raises(NodesConnectionError):
        subgraph_1 >> node_d


def test__cyclic_connection_error(node_a):
    with pytest.raises(CyclicNodeError):
        node_a >> node_a


def test__nodes_already_connected_error(node_a, node_b):
    with pytest.raises(NodesAlreadyConnectedError):
        node_a >> node_b
        node_a >> node_b


def test__node_input_already_connected_error(node_a, node_b, node_c):
    with pytest.raises(NodesConnectionError):
        node_a >> node_b
        _ = node_c >> node_b



from dataclasses import dataclass, field
from typing import List, Set, Union, Dict, Tuple

from cupyd.core.communication import InterProcessConnector
from cupyd.core.computing.etl_worker import ETLWorkerThread, ETLWorkerProcess
from cupyd.core.graph.classes import Node


@dataclass
class ETLSegment:
    id: str
    nodes: List[Node]
    node_ids: Set[str]
    run_in_main_process: bool
    num_workers: int

    # these will be filled when building the ETL
    workers_by_id: Dict[str, Union[ETLWorkerThread, ETLWorkerProcess]] = field(
        default_factory=lambda: dict()
    )
    # first tuple element is the amount of ETLWorkers that are consuming from the Connector
    output_interprocess_connectors: List[Tuple[int, InterProcessConnector]] = field(
        default_factory=lambda: []
    )



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Iterator, final

from cupyd.core.graph.classes import Node
from cupyd.core.models.node_configuration import ExtractorConfiguration


class Extractor(Node):
    def __init__(self):
        super().__init__()
        self._configuration = ExtractorConfiguration()

    @abstractmethod
    def extract(self) -> Iterator[Any]:
        """Extract an item."""

        raise NotImplementedError("Missing implementation of extract() method!")

    @final
    @property
    def configuration(self):
        return self._configuration



import logging
from collections import deque
from itertools import chain
from multiprocessing import Queue
from statistics import median
from threading import Thread
from time import time
from typing import Optional, List

from cupyd.core.communication import InterProcessEventFlag
from cupyd.core.constants.logging import LOGGING_MSG_PADDING
from cupyd.core.constants.sentinel_values import NO_MORE_ITEMS
from cupyd.core.graph.classes import Node

logger = logging.getLogger("cupyd.timings")


class TimingsThread(Thread):

    def __init__(
        self,
        nodes: List[Node],
        node_timings: Queue,
        stop_event: InterProcessEventFlag,
        refresh_interval: int = 5,  # seconds
    ):
        super().__init__(name="cupyd")
        self.node_timings = node_timings
        self.stop_event = stop_event
        self.refresh_interval = refresh_interval
        self.node_name_by_id = {node.id: node.name for node in nodes}
        self.stats_by_node_id = {
            node.id: {"min": None, "max": None, "median": None} for node in nodes
        }
        self.buffer_by_node_id = {node.id: deque(maxlen=100) for node in nodes}

    def run(self):
        last_log_time = time()

        while True:
            item = self.node_timings.get()

            if item is NO_MORE_ITEMS:
                break
            elif self.stop_event:
                continue
            else:
                node_id, timing = item

            self.buffer_by_node_id[node_id].append(timing)

            current_time = time()

            if (last_log_time - current_time) >= self.refresh_interval:
                last_log_time = current_time
                self._update_timings()
                self._log_timings()

        if not self.stop_event:
            self._update_timings()
            self._log_timings()

    def _update_timings(self):
        for node_id, timings in self.buffer_by_node_id.items():
            if timings:
                self.stats_by_node_id[node_id]["median"] = median(timings)

                if self.stats_by_node_id[node_id]["min"]:
                    current_min = [self.stats_by_node_id[node_id]["min"]]
                    current_max = [self.stats_by_node_id[node_id]["max"]]
                else:
                    current_min = []
                    current_max = []

                self.stats_by_node_id[node_id]["min"] = min(chain(timings, current_min))
                self.stats_by_node_id[node_id]["max"] = max(chain(timings, current_max))

    def _log_timings(self):
        log = "timings:\n"
        for node_id, stats in self.stats_by_node_id.items():
            min_timing = self._format_timing(stats["min"])
            max_timing = self._format_timing(stats["max"])
            median_timing = self._format_timing(stats["median"])

            log += (
                f"{LOGGING_MSG_PADDING}\t• {self.node_name_by_id[node_id]}\n"
                f"{LOGGING_MSG_PADDING}\t\t"
                f"{median_timing} (avg) | {min_timing} (min) | {max_timing} (max)\n"
            )

        logger.info(log)

    @staticmethod
    def _format_timing(value: Optional[float]) -> str:
        if value is None:
            return "no_measure"

        if 1 <= value < 60:
            return f"{round(value, 2)} s"
        elif value < 1:
            ms_value = value * 1000
            if ms_value < 1:
                return f"{round(ms_value * 1000, 4)} μs"
            else:
                return f"{round(ms_value, 4)} ms"
        else:
            minutes_value = value / 60
            if minutes_value <= 1:
                return f"{round(minutes_value, 2)} min"
            else:
                return f"{round(minutes_value / 60, 4)} hours"



"""In this example we will compute the factorial of 50.000 integers, using multiprocessing,
while storing the results into 2 separate lists, one of even results and another for odd ones."""

import math
import logging
from time import time
from typing import Any

from cupyd import ETL, Extractor, Transformer, Loader, Filter

logger = logging.getLogger("readme_etl")


class IntegerExtractor(Extractor):

    def __init__(self, total_items: int):
        super().__init__()
        self.total_items = total_items

        # generated integers will be passed onto each worker in buckets of size 10
        self.configuration.bucket_size = 10

    def extract(self) -> int:
        for item in range(self.total_items):
            yield item


class Factorial(Transformer):

    def transform(self, item: int) -> int:
        return math.factorial(item)


class EvenOnly(Filter):

    def filter(self, item: int) -> int | None:
        return item if item & 1 else None


class OddOnly(Filter):

    def filter(self, item: int) -> int | None:
        return None if item & 1 else item


class ListLoader(Loader):

    def __init__(self):
        super().__init__()
        self.configuration.run_in_main_process = True
        self.items = []

    def start(self):
        self.items = []

    def load(self, item: Any):
        self.items.append(item)


def compute_with_single_process():
    start_time = time()

    even_results = []
    odd_results = []

    logger.info("Running README script without cupyd...")

    for value in range(50_000):
        result = math.factorial(value)
        if result & 1:
            even_results.append(factorial)
        else:
            odd_results.append(factorial)

    logger.info(f"Finished! Elapsed time: {time() - start_time} seconds")


if __name__ == "__main__":
    # 1. Instantiate the ETL Nodes
    ext = IntegerExtractor(total_items=100)
    factorial = Factorial()
    even_only = EvenOnly()
    odd_only = OddOnly()
    even_ldr = ListLoader()
    odd_ldr = ListLoader()

    # 2. Connect the Nodes to determine the data flow. Notice the two ETL branches after the
    # factorial is computed
    ext >> factorial >> [even_only >> even_ldr, odd_only >> odd_ldr]

    # 3. Run the ETL with 8 workers (multiprocessing Processes)
    etl = ETL(extractor=ext)
    etl.run(workers=8, show_progress=True, monitor_performance=True)

    # 4. You can access the results stored in both Loaders after the ETL is finished
    even_factorials = even_ldr.items
    odd_factorials = odd_ldr.items

    # Let's compare the performance with a single core, although without the cupyd overhead
    compute_with_single_process()



import logging
from collections import defaultdict
from copy import deepcopy
from multiprocessing import Queue, set_start_method, get_start_method
from time import time
from typing import List, Dict, Optional

from cupyd.core.communication.connector import (
    Connector,
    IntraProcessConnector,
    InterProcessConnector,
)
from cupyd.core.communication.counter import MPCounter
from cupyd.core.communication.event_flag import (
    InterProcessEventFlag,
    IntraProcessEventFlag,
)
from cupyd.core.communication.interruption_handler import InterruptionHandler
from cupyd.core.computing.etl_worker import ETLWorkerProcess, ETLWorkerThread
from cupyd.core.constants.logging import LOGGING_FORMAT
from cupyd.core.constants.sentinel_values import NO_MORE_ITEMS
from cupyd.core.exceptions import ETLExecutionError, InterruptedETL
from cupyd.core.graph.algorithms import (
    get_etl_segments,
    topological_sort,
    assign_names_and_ids_to_nodes,
)
from cupyd.core.graph.classes import Node
from cupyd.core.models.etl_segment import ETLSegment
from cupyd.core.models.node_exception import NodeException
from cupyd.core.nodes import Loader
from cupyd.core.nodes.extractor import Extractor
from cupyd.core.stats.progress_thread import ProgressThread
from cupyd.core.stats.timings_thread import TimingsThread
from cupyd.core.utils import format_seconds, get_subdict

logger = logging.getLogger("cupyd.etl")


class ETL:

    def __init__(self, extractor: Extractor):
        if not isinstance(extractor, Extractor):
            raise TypeError(f"Invalid Extractor: {type(extractor)}")
        self.extractor = extractor

    def run(
        self,
        workers: int = 1,
        raise_exception: bool = True,
        raise_exception_if_interrupted: bool = True,
        monitor_performance: bool = False,
        show_progress: bool = True,
        verbose: bool = True,
    ):
        start_method = get_start_method()
        set_start_method("spawn", force=True)  # ETLWorkerProcesses will be spawned

        current_formatter, current_level, current_logfile = self._setup_logger()

        (
            nodes,
            segments_by_id,
            stop_event,
            pause_event,
            interruption_handler,
            node_timings,
            finished_workers,
            monitor_performance_event_by_node_id,
            counter_by_node_id,
        ) = self._build(num_workers=workers, monitor_performance=monitor_performance)

        if verbose:
            logger.info("ETL build successful, running ETL...")

        start_time = time()
        exceptions_by_node_id: Dict[str, List[NodeException]] = defaultdict(list)

        # this will handle interruptions in this main process (and the ETLWorkerThreads)
        interruption_handler.start()

        if monitor_performance:
            timings_thread = TimingsThread(
                nodes=nodes, node_timings=node_timings, stop_event=stop_event
            )
            timings_thread.start()
        else:
            timings_thread = None

        if show_progress:
            finalize_event = IntraProcessEventFlag()
            progress_thread = ProgressThread(
                nodes=nodes,
                counter_by_node_id=counter_by_node_id,
                stop_event=stop_event,
                finalize_event=finalize_event,
            )
            progress_thread.start()
        else:
            progress_thread, finalize_event = None, None

        # start all ETLWorkers
        for segment in segments_by_id.values():
            for worker in segment.workers_by_id.values():
                worker.start()

        # auxiliary structures to easily access segment resources
        active_worker_ids_by_segment_id = {}
        workers_by_id = {}

        for segment_id, segment in segments_by_id.items():
            active_worker_ids_by_segment_id[segment_id] = set(segment.workers_by_id.keys())

            for worker_id, worker in segment.workers_by_id.items():
                workers_by_id[worker_id] = worker

        if verbose:
            logger.info(f"ETL startup time: {round(time() - start_time, 4)} seconds")

        # run until all ETLSegments are finished
        while active_worker_ids_by_segment_id:
            try:
                worker_id, segment_id, exception_by_node_id = finished_workers.get()
                workers_by_id[worker_id].join()
                active_worker_ids_by_segment_id[segment_id].remove(worker_id)

                if exception_by_node_id:
                    for node_id, exception in exception_by_node_id.items():
                        exceptions_by_node_id[node_id].append(exception)

                # if whole segment has finished, we will trigger the finalization of other connected
                # nodes (from other segments if connected via InterProcessConnector)
                if not active_worker_ids_by_segment_id[segment_id]:
                    active_worker_ids_by_segment_id.pop(segment_id)

                    for output_segment_num_workers, connector in segments_by_id[
                        segment_id
                    ].output_interprocess_connectors:
                        for _ in range(output_segment_num_workers):
                            connector.produce(NO_MORE_ITEMS)
            except InterruptedError:
                continue

        # stop the TimingsThread & ProgressThread, if running
        if monitor_performance:
            node_timings.put(NO_MORE_ITEMS)
            timings_thread.join()

        if show_progress:
            finalize_event.set()
            progress_thread.join()

        # close InterProcessConnectors
        for segment in segments_by_id.values():
            for _, connector in segment.output_interprocess_connectors:
                connector.close()

        elapsed_time = format_seconds(time() - start_time)

        # restore multiprocessing start method & the previous signal handlers
        set_start_method(start_method, force=True)
        interruption_handler.restore_handlers()

        if exceptions_by_node_id:
            failed_nodes_names = []
            for node in nodes:
                if node.id in exceptions_by_node_id:
                    failed_nodes_names.append(node.name)
            failed_nodes_names.sort()
            logger.error(
                f'ETL finished with errors in nodes: {", ".join(failed_nodes_names)} | '
                f"Elapsed time: {elapsed_time}"
            )
        else:
            if interruption_handler.interrupted():
                if raise_exception_if_interrupted:
                    raise InterruptedETL(f"Elapsed time: {elapsed_time}")
                else:
                    logger.warning(f"ETL interrupted! | {elapsed_time}")
            else:
                logger.info(f"ETL finished | Elapsed time: {elapsed_time}")

        self._reset_root_logging(
            current_level=current_level,
            current_formatter=current_formatter,
            current_logfile=current_logfile,
        )

        if exceptions_by_node_id and raise_exception:
            # we will only raise one exception although maybe multiple occurred
            for exceptions in exceptions_by_node_id.values():
                raise ETLExecutionError(exceptions[0].traceback_formatted)

    def _build(self, num_workers: int, monitor_performance: bool):
        """Build the ETL."""

        # 1. List all nodes & edges & ensure the ETL is a DAG
        nodes, edges = topological_sort(root_node=self.extractor)

        # 2. Each node will have its unique ID + an autogenerated name (if any wasn't provided)
        assign_names_and_ids_to_nodes(nodes=nodes)

        # 3. Determine the ETLSegments of the ETL. Each ETLSegment will consist in a group of
        # consecutive nodes that will run on the same ETLWorker
        segments = get_etl_segments(nodes=nodes, num_workers=num_workers)

        # 4. Create the ETL connectors
        input_connector_by_node_id: Dict[str, Connector] = {}
        output_connectors_by_node_id: Dict[str, List[Connector]] = defaultdict(list)

        for edge in edges:
            origin: Node = edge.origin
            target: Node = edge.target

            origin_segment = None
            target_segment = None

            for segment in segments:
                if origin.id in segment.node_ids:
                    origin_segment = segment
                if target.id in segment.node_ids:
                    target_segment = segment

            # TODO: allow customizing these max sizes when configuring the ETL/Nodes
            if origin_segment.id == target_segment.id:
                connector = IntraProcessConnector(maxsize=0)
            else:
                connector = InterProcessConnector(maxsize=0)
                connector.start()

            input_connector_by_node_id[target.id] = connector
            output_connectors_by_node_id[origin.id].append(connector)

            if isinstance(connector, InterProcessConnector):
                origin_segment.output_interprocess_connectors.append(
                    (target_segment.num_workers, connector)
                )

        # 5. Create necessary structures & ETL Workers
        worker_num = 1
        finished_workers = Queue()
        node_timings = Queue(maxsize=100_000)  # todo: might cause slowdowns, rethink?
        stop_event = InterProcessEventFlag()
        pause_event = InterProcessEventFlag()
        interruption_handler = InterruptionHandler(stop_event=stop_event)
        counter_by_node_id: Dict[str, MPCounter] = {
            node.id: MPCounter() for node in nodes if isinstance(node, Loader) and not node.outputs
        }
        monitor_performance_event_by_node_id = {
            node.id: InterProcessEventFlag(start_set=monitor_performance) for node in nodes
        }
        segments_by_id: Dict[str, ETLSegment] = {}

        for segment in segments:
            worker_class = ETLWorkerThread if segment.run_in_main_process else ETLWorkerProcess

            for _ in range(segment.num_workers):
                etl_worker_id = f"etl_worker_{worker_num}"

                etl_worker = worker_class(
                    worker_id=etl_worker_id,
                    segment_id=segment.id,
                    nodes=segment.nodes,
                    counters=get_subdict(dictionary=counter_by_node_id, keys=segment.node_ids),
                    input_connector_by_node_id=get_subdict(
                        dictionary=input_connector_by_node_id, keys=segment.node_ids
                    ),
                    output_connectors_by_node_id=get_subdict(
                        dictionary=output_connectors_by_node_id, keys=segment.node_ids
                    ),
                    monitor_performance_event_by_node_id=get_subdict(
                        dictionary=monitor_performance_event_by_node_id,
                        keys=segment.node_ids,
                    ),
                    stop_event=stop_event,
                    pause_event=pause_event,
                    interruption_handler=interruption_handler,
                    node_timings=node_timings,
                    finished_workers=finished_workers,
                )
                segment.workers_by_id[etl_worker_id] = etl_worker
                worker_num += 1

            segments_by_id[segment.id] = segment

        return (
            nodes,
            segments_by_id,
            stop_event,
            pause_event,
            interruption_handler,
            node_timings,
            finished_workers,
            monitor_performance_event_by_node_id,
            counter_by_node_id,
        )

    @staticmethod
    def _setup_logger() -> [logging.Formatter, int, Optional[str]]:
        """Update the current logger handler & backup the current logging configuration."""

        try:
            current_formatter = deepcopy(logging.root.handlers[0].formatter)
        except Exception:
            logger.warning("Unable to get current logger")
            return None, None, None

        current_level = logging.root.level

        try:
            current_logfile = logging.root.handlers[0].baseFilename
        except AttributeError:
            current_logfile = None

        if current_logfile:
            logger_handler = logging.FileHandler(filename=current_logfile)
        else:
            logger_handler = logging.StreamHandler()

        logger_handler.setLevel(logging.INFO)
        logger_handler.setFormatter(logging.Formatter(LOGGING_FORMAT))
        logger.propagate = False
        logger.addHandler(logger_handler)

        return current_formatter, current_level, current_logfile

    @staticmethod
    def _reset_root_logging(
        current_level: Optional[int],
        current_formatter: Optional[logging.Formatter],
        current_logfile: Optional[str],
    ):
        """Restore the old logging configuration."""

        if not current_formatter:
            return None

        logging.root.setLevel(current_level)
        for idx, h in enumerate(logging.root.handlers):
            if current_logfile:
                handler = logging.FileHandler(filename=current_logfile)
                handler.setFormatter(current_formatter)
                logging.root.handlers[idx] = handler
            else:
                h.setFormatter(current_formatter)



from dataclasses import dataclass
from typing import Any


@dataclass
class ExtractorConfiguration:
    bucket_size: int = 1000
    run_in_main_process: bool = True


@dataclass
class TransformerConfiguration:
    input_key: str = None
    run_in_main_process: bool = False


@dataclass
class FilterConfiguration:
    input_key: str = None
    value_to_filter: Any = None
    disable_safe_copy: bool = False
    run_in_main_process: bool = False


@dataclass
class LoaderConfiguration:
    input_key: str = None
    disable_safe_copy: bool = False
    run_in_main_process: bool = False


@dataclass
class BulkerConfiguration:
    run_in_main_process: bool = False


class DeBulkerConfiguration:
    run_in_main_process: bool = False



from multiprocessing import RawValue, Lock


class MPCounter:
    def __init__(self):
        self._value = RawValue("i", 0)
        self._lock = Lock()

    def increase(self, amount: int):
        with self._lock:
            self._value.value += amount

    @property
    def value(self):
        return self._value.value



from typing import Any, Iterator, List, Optional

from cupyd import Transformer, Loader, Extractor, Filter


class ListExtractor(Extractor):

    def __init__(self, items: List[Any]):
        super().__init__()
        self.configuration.run_in_main_process = True
        self.items = items

    def extract(self) -> Iterator[Any]:
        for item in self.items:
            yield item


class AdderToStr(Transformer):

    def transform(self, item: int) -> str:
        item += 5
        return str(item)


class CustomFilter(Filter):

    def filter(self, item: str) -> Optional[str]:
        if int(item) % 5 == 0:
            return None
        else:
            return item


class ListLoader(Loader):
    def __init__(self):
        super().__init__()
        self.configuration.run_in_main_process = True
        self.items = []

    def start(self):
        self.items = []

    def load(self, item: Any):
        self.items.append(item)



from __future__ import annotations

from typing import final

from cupyd.core.graph.classes import Node
from cupyd.core.models.node_configuration import DeBulkerConfiguration


class DeBulker(Node):
    def __init__(self):
        super().__init__()
        self._configuration = DeBulkerConfiguration()

    @final
    @property
    def configuration(self):
        return self._configuration



from multiprocessing import SimpleQueue, Value
from typing import List


# TODO: Not used, tentative to optimize the timings part
class TimingsBuffer:

    def __init__(self, maxsize: int = 25):
        super().__init__()
        self._maxsize = maxsize
        self._queue = SimpleQueue()
        self._buffer_size = Value("i", 0)

    def produce_timing(self, timing: float):
        with self._buffer_size.get_lock():
            if self._buffer_size.value >= self._maxsize:
                self._queue.get()  # make space in the buffer with the new timing
            else:
                self._buffer_size.value += 1
            self._queue.put(timing)

    def consume_buffer(self) -> List[float]:
        with self._buffer_size.get_lock():
            buffer_size = self._buffer_size.value
            if buffer_size == 0:
                timings = []
            else:
                timings = [self._queue.get() for _ in range(buffer_size)]
                self._buffer_size.value = 0

            return timings

    def close(self):
        self._queue.close()



from abc import abstractmethod
from copy import deepcopy
from functools import partial
from multiprocessing import Queue as MultiprocessingQueue
from queue import Queue
from threading import Thread
from time import perf_counter
from typing import List, Optional, Union, Any, Iterator

from cupyd.core.communication import Connector, InterProcessEventFlag
from cupyd.core.communication.connector import IntraProcessConnector
from cupyd.core.communication.counter import MPCounter
from cupyd.core.constants.node_actions import (
    START,
    FINALIZE,
    UPDATE_COUNTER,
    PRODUCE_BUCKET,
    PRODUCE_TIMING,
    PROCESS_BUCKET,
    CONSUME_BUCKET,
    GENERATE_BUCKET,
)
from cupyd.core.constants.sentinel_values import NO_MORE_ITEMS
from cupyd.core.graph.classes import Node
from cupyd.core.models.node_exception import NodeException
from cupyd.core.nodes.bulker import Bulker
from cupyd.core.nodes.debulker import DeBulker
from cupyd.core.nodes.extractor import Extractor
from cupyd.core.nodes.filter import Filter
from cupyd.core.nodes.loader import Loader
from cupyd.core.nodes.transformer import Transformer


def _set_copy_bucket_on_produce(output_connectors: List[Connector]) -> List[Connector]:
    """Determine which connectors need a pre-copy of the bucket before producing them."""

    intra_process_connector_found = False

    for connector in output_connectors:
        if isinstance(connector, IntraProcessConnector):
            if intra_process_connector_found:
                connector.copy_bucket_on_produce = True
            else:
                intra_process_connector_found = True

    return output_connectors


def get_item_value_by_key(item: Any, key: str) -> Any:
    if isinstance(item, dict):
        return item[key]
    else:
        return getattr(item, key)


def _extractor_item_generator(extractor: Extractor) -> Iterator[List[Any]]:
    for item in extractor.extract():
        yield item


def _transformer_process_bucket(
    bucket: List[Any],
    transformer: Transformer,
    input_key: Optional[str],
) -> List[Any]:
    """Run the transform() method on every item from a bucket."""

    processed_bucket = []

    for item in bucket:

        if input_key:
            value = get_item_value_by_key(item=item, key=input_key)
            result = transformer.transform(value)
        else:
            result = transformer.transform(item)

        processed_bucket.append(result)

    return processed_bucket


def _filter_process_bucket(
    bucket: List[Any],
    filter_node: Filter,
    value_to_filter: Any,
    disable_safe_copy: bool,
    input_key: Optional[str],
) -> List[Any]:
    """Run the filter() method on every item from a bucket."""

    filtered_bucket = []

    items = bucket if disable_safe_copy else deepcopy(bucket)

    for item_idx, item in enumerate(items):

        if input_key:
            value = get_item_value_by_key(item=item, key=input_key)
            result = filter_node.filter(value)
        else:
            result = filter_node.filter(item)

        if result != value_to_filter:
            filtered_bucket.append(bucket[item_idx])

    return filtered_bucket


def _loader_process_bucket(
    bucket: List[Any],
    loader: Loader,
    has_outputs: bool,
    disable_safe_copy: bool,
    input_key: Optional[str],
) -> List[Any]:
    """Run the load() method on every item from a bucket."""

    # If the Loader output items are passed onto another Node, then, as a safety measure (unless
    # disabled), it will load a copy of the incoming bucket instead of the original one
    if has_outputs:
        items = bucket if disable_safe_copy else deepcopy(bucket)
        for item in items:
            if input_key:
                value = get_item_value_by_key(item=item, key=input_key)
                loader.load(value)
            else:
                loader.load(item)
    else:
        for item in bucket:
            if input_key:
                value = get_item_value_by_key(item=item, key=input_key)
                loader.load(value)
            else:
                loader.load(item)

    return bucket


class NodeWorker(Thread):

    def __init__(
        self,
        node: Union[Node, Extractor, Transformer, Filter, Loader, Bulker, DeBulker],
        input_connector: Optional[Connector],
        output_connectors: List[Connector],
        counter: Optional[MPCounter],
        finished_threads_queue: Queue,
        stop_event: InterProcessEventFlag,
        pause_event: InterProcessEventFlag,
        monitor_performance: InterProcessEventFlag,
        node_timings: MultiprocessingQueue,
    ):
        super().__init__(name=node.name)
        self.node = node
        self.input_connector = input_connector
        self.output_connectors = output_connectors
        self.counter = counter
        self.finished_threads_queue = finished_threads_queue
        self.stop_event = stop_event
        self.pause_event = pause_event
        self.node_timings = node_timings
        self.monitor_performance = monitor_performance
        self.exception_found: Optional[NodeException] = None
        self.skip_processing = False
        self.is_node_terminal = isinstance(self.node, Loader) and not self.node.outputs

    def run(self):
        # todo: could this be determined beforehand? should be possible, at build() step
        self.output_connectors = _set_copy_bucket_on_produce(self.output_connectors)

        if not isinstance(self.node, Bulker) and not isinstance(self.node, DeBulker):
            try:
                self.node.start()
            except Exception as exc:
                self.exception_found = NodeException(exc=exc, action=START)

        if not self.exception_found:
            self._run()

        if not isinstance(self.node, Bulker) and not isinstance(self.node, DeBulker):
            try:
                if self.exception_found:
                    self.node.handle_exception(self.exception_found)
                else:
                    self.node.finalize()
            except Exception as e:
                self.exception_found = NodeException(exc=e, action=FINALIZE)

        self.finished_threads_queue.put((self.node.id, self.exception_found))

    @abstractmethod
    def _run(self):
        pass

    def _handle_exception(self, exception: Exception, action: str) -> None:
        if not self.stop_event:
            self.stop_event.set()
        if not self.exception_found:  # don't replace original exception in case another one occurs
            self.exception_found = NodeException(exc=exception, action=action)
        self.skip_processing = True


class ExtractorWorker(NodeWorker):

    def _run(self):
        self.node: Extractor

        bucket = []
        stop_iteration = False
        start_time: Optional[float] = None
        bucket_size = self.node.configuration.bucket_size
        generator = _extractor_item_generator(extractor=self.node)

        while not self.stop_event:

            if self.pause_event:
                self.pause_event.wait()

            # generate a bucket of items
            try:
                start_time = perf_counter() if self.monitor_performance else None

                while len(bucket) < bucket_size:
                    bucket.append(next(generator))

            except StopIteration:
                stop_iteration = True
            except Exception as e:
                self.exception_found = NodeException(exc=e, action=GENERATE_BUCKET)
                break

            # produce the bucket to the output connectors of the node
            if bucket:
                try:
                    for connector in self.output_connectors:
                        connector.produce(bucket)
                except Exception as e:
                    self.exception_found = NodeException(exc=e, action=PRODUCE_BUCKET)
                    break

                try:
                    if start_time:
                        timing = (perf_counter() - start_time) / len(bucket)
                        self.node_timings.put((self.node.id, timing))
                except Exception as e:
                    self.exception_found = NodeException(exc=e, action=PRODUCE_TIMING)
                    break

                bucket = []

            if stop_iteration:
                break


class ProcessorWorker(NodeWorker):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if isinstance(self.node, Transformer):
            self._process_bucket_function = partial(
                _transformer_process_bucket,
                transformer=self.node,
                input_key=self.node.configuration.input_key,
            )
        elif isinstance(self.node, Loader):
            self._process_bucket_function = partial(
                _loader_process_bucket,
                loader=self.node,
                has_outputs=bool(self.node.outputs),
                disable_safe_copy=self.node.configuration.disable_safe_copy,
                input_key=self.node.configuration.input_key,
            )
        elif isinstance(self.node, Filter):
            self._process_bucket_function = partial(
                _filter_process_bucket,
                filter_node=self.node,
                value_to_filter=self.node.configuration.value_to_filter,
                disable_safe_copy=self.node.configuration.disable_safe_copy,
                input_key=self.node.configuration.input_key,
            )
        else:
            raise TypeError(f"Invalid node type: {type(self.node)}")

    # TODO: Might be useful to know which item value caused the error
    def _run(self):
        timing: Optional[float] = None
        bucket = None

        while True:
            # consume a bucket of items
            try:
                bucket = self.input_connector.consume()
                if bucket is NO_MORE_ITEMS:
                    break
            except Exception as e:
                self._handle_exception(exception=e, action=CONSUME_BUCKET)

            if self.stop_event:
                self.skip_processing = True

            if self.skip_processing:
                continue

            if self.pause_event:
                self.pause_event.wait()

            # process the bucket
            # todo: detect the stop or pause event as soon its set, checking on every item
            #  instead on every bucket? Add timeout to avoid hanging processing after stop was set?
            try:
                start_time = perf_counter() if self.monitor_performance else None

                bucket = self._process_bucket_function(bucket)

                if start_time and bucket:
                    timing = (perf_counter() - start_time) / len(bucket)

            except Exception as e:
                self._handle_exception(exception=e, action=PROCESS_BUCKET)
            else:
                # produce the bucket to the output connectors of the node
                try:
                    for connector in self.output_connectors:
                        connector.produce(bucket)
                except Exception as e:
                    self._handle_exception(exception=e, action=PRODUCE_BUCKET)
                    continue

                try:
                    if timing:
                        self.node_timings.put((self.node.id, timing))
                        timing = None
                except Exception as e:
                    self._handle_exception(exception=e, action=PRODUCE_TIMING)
                    continue

                try:
                    if self.counter:
                        self.counter.increase(amount=len(bucket))
                except Exception as e:
                    self._handle_exception(exception=e, action=UPDATE_COUNTER)
                    continue


class BulkerWorker(NodeWorker):

    def _run(self):
        bulk = []
        bulk_size = self.node.bulk_size  # TODO: allow changing bulk size dynamically?

        while True:

            # consume a bucket of items
            try:
                bucket = self.input_connector.consume()
                if bucket is NO_MORE_ITEMS:
                    break
            except Exception as e:
                self._handle_exception(exception=e, action=CONSUME_BUCKET)
                continue

            if self.stop_event:
                self.skip_processing = True

            if self.skip_processing:
                continue

            if self.pause_event:
                self.pause_event.wait()

            try:
                bulk.extend(bucket)

                if len(bulk) >= bulk_size:
                    remaining_items = None

                    for chunk in self._chunk(items=bulk, bulk_size=bulk_size):
                        if len(chunk) == bulk_size:
                            for connector in self.output_connectors:
                                connector.produce([chunk])
                        else:
                            remaining_items = chunk

                    bulk = remaining_items or []

            except Exception as e:
                self._handle_exception(exception=e, action=PRODUCE_BUCKET)
                continue

        # if there are remaining items, produce them
        if bulk and self.exception_found is None:
            try:
                for connector in self.output_connectors:
                    connector.produce([bulk])
            except Exception as e:
                self._handle_exception(exception=e, action=PRODUCE_BUCKET)

    @staticmethod
    def _chunk(items: List[Any], bulk_size: int) -> List[Any]:
        for i in range(0, len(items), bulk_size):
            yield items[i : i + bulk_size]  # noqa


class DeBulkerWorker(NodeWorker):

    def _run(self):
        while True:

            # consume a bucket of items
            try:
                bucket = self.input_connector.consume()
                if bucket is NO_MORE_ITEMS:
                    break
            except Exception as e:
                self._handle_exception(exception=e, action=CONSUME_BUCKET)
                continue

            if self.stop_event:
                self.skip_processing = True

            if self.skip_processing:
                continue

            if self.pause_event:
                self.pause_event.wait()

            try:
                for item in bucket:
                    for connector in self.output_connectors:
                        connector.produce(item)
            except Exception as e:
                self._handle_exception(exception=e, action=PRODUCE_BUCKET)
                continue



from datetime import timedelta
from typing import Union, Dict, Iterable


def format_seconds(seconds: Union[int, float]) -> str:
    return str(timedelta(seconds=seconds))


def get_subdict(dictionary: Dict, keys: Iterable[str]) -> Dict:
    return {key: dictionary[key] for key in keys if key in dictionary}



import logging
from threading import Thread
from time import sleep, time
from typing import List, Dict, Optional

from cupyd.core.communication.counter import MPCounter
from cupyd.core.communication.event_flag import (
    IntraProcessEventFlag,
    InterProcessEventFlag,
)
from cupyd.core.constants.logging import LOGGING_MSG_PADDING
from cupyd.core.graph.classes import Node
from cupyd.core.utils import format_seconds

logger = logging.getLogger("cupyd.progress")


class ProgressThread(Thread):

    def __init__(
        self,
        nodes: List[Node],
        counter_by_node_id: Dict[str, MPCounter],
        finalize_event: IntraProcessEventFlag,
        stop_event: InterProcessEventFlag,
        refresh_interval: float = 2.5,  # seconds
    ):
        super().__init__(name="cupyd (progress)")
        self.finalize_event = finalize_event
        self.stop_event = stop_event
        self.refresh_interval = refresh_interval
        self.counter_by_node_name = {}
        self.start_time: Optional[float] = None

        for node in nodes:
            if node.id in counter_by_node_id:
                self.counter_by_node_name[node.name] = counter_by_node_id[node.id]

    def run(self):
        self.start_time = time()
        last_state = self._get_counters_state()

        while not self.finalize_event and not self.stop_event:
            new_state = self._get_counters_state()
            if new_state != last_state:
                last_state = new_state
                self._log_progress()
            sleep(self.refresh_interval)

        if not self.stop_event:
            self._log_progress(last_log=True)

    def _get_counters_state(self):
        return {
            node_name: int(counter.value)
            for node_name, counter in self.counter_by_node_name.items()
        }

    def _log_progress(self, last_log: bool = False):
        if last_log:
            log = "[FINISHED] progress:\n"
        else:
            log = f"ET: {format_seconds(time() - self.start_time)} | Loaded items:\n"

        for node_name, counter in self.counter_by_node_name.items():
            total_items = f"{counter.value:,}"
            log += f"{LOGGING_MSG_PADDING}\t• {node_name}: {total_items}\n"
        logger.info(log)



from unittest import TestCase

from cupyd import ETL
from cupyd.tests.etl.nodes import ListExtractor, AdderToStr, ListLoader, CustomFilter


def test__etl_01():
    test_case = TestCase()

    items = [0, 1, 2, 3, 4, 5]
    expected_items = ["5", "6", "7", "8", "9", "10"]

    ext = ListExtractor(items)
    tf = AdderToStr()
    ldr = ListLoader()

    ext >> tf >> ldr
    ETL(ext).run()

    test_case.assertListEqual(ldr.items, expected_items)


def test__etl_02():
    test_case = TestCase()

    items = [0, 1, 2, 3, 4, 5]
    expected_items = ["5", "6", "7", "8", "9", "10"]

    ext = ListExtractor(items)
    tf = AdderToStr()
    ldr_1 = ListLoader()
    ldr_2 = ListLoader()

    ext >> tf >> [ldr_1, ldr_2]
    ETL(ext).run()

    test_case.assertListEqual(ldr_1.items, expected_items)
    test_case.assertListEqual(ldr_2.items, expected_items)


def test__etl_03():
    test_case = TestCase()

    items = [0, 1, 2, 3, 4, 5]
    expected_items = ["5", "6", "7", "8", "9", "10"]

    ext = ListExtractor(items=items)
    tf = AdderToStr()
    ldr_1 = ListLoader()
    ldr_2 = ListLoader()

    ext >> tf >> [ldr_1, ldr_2]
    ETL(ext).run()

    test_case.assertCountEqual(ldr_1.items, expected_items)
    test_case.assertCountEqual(ldr_2.items, expected_items)


def test__etl_04():
    test_case = TestCase()

    items = [0, 1, 2, 3, 4, 5]
    expected_items_1 = ["5", "6", "7", "8", "9", "10"]
    expected_items_2 = ["6", "7", "8", "9"]

    ext = ListExtractor(items=items)
    tf = AdderToStr()
    filter_1 = CustomFilter()
    ldr_1 = ListLoader()
    ldr_2 = ListLoader()

    ext >> tf >> [ldr_1, filter_1 >> ldr_2]
    ETL(ext).run()

    test_case.assertCountEqual(ldr_1.items, expected_items_1)
    test_case.assertCountEqual(ldr_2.items, expected_items_2)



START = "START"
FINALIZE = "FINALIZE"
PROCESS_BUCKET = "PROCESS_BUCKET"
GENERATE_BUCKET = "GENERATE_BUCKET"
CONSUME_BUCKET = "CONSUME_BUCKET"
PRODUCE_BUCKET = "PRODUCE_BUCKET"
PRODUCE_TIMING = "PRODUCE_TIMING"
UPDATE_COUNTER = "UPDATE_COUNTER"

NODE_ACTIONS = [
    START,
    FINALIZE,
    CONSUME_BUCKET,
    PRODUCE_BUCKET,
    PROCESS_BUCKET,
    GENERATE_BUCKET,
    PRODUCE_TIMING,
    UPDATE_COUNTER,
]



import ctypes
import logging
import signal
from multiprocessing import Lock, Value

from cupyd.core.communication.event_flag import InterProcessEventFlag

logger = logging.getLogger("cupyd.interrupt")


class InterruptionHandler:
    """Class responsible for setting up a handler for termination signals, in order to perform
    clean shutdowns of running ETLs."""

    def __init__(self, stop_event: InterProcessEventFlag):
        self._lock = Lock()
        self._stop_event = stop_event
        self._interrupted = Value(ctypes.c_bool, False)
        self._original_signal_handlers = {}

    def start(self):
        self._original_signal_handlers[signal.SIGINT] = signal.signal(
            signal.SIGINT, self._handle_signal
        )
        self._original_signal_handlers[signal.SIGTERM] = signal.signal(
            signal.SIGTERM, self._handle_signal
        )
        try:
            self._original_signal_handlers[signal.SIGQUIT] = signal.signal(
                signal.SIGQUIT, self._handle_signal
            )
            self._original_signal_handlers[signal.SIGHUP] = signal.signal(
                signal.SIGHUP, self._handle_signal
            )
        except AttributeError:
            pass

    def interrupted(self) -> bool:
        with self._lock:
            return self._interrupted.value

    def _handle_signal(self, *_):
        with self._lock:
            interrupted = self._interrupted.value
            if not interrupted:
                logger.warning("Interruption signal detected, stopping the ETL...")
                self._interrupted.value = True
                if not self._stop_event:
                    self._stop_event.set()

    def restore_handlers(self):
        for signal_, original_signal_handler in self._original_signal_handlers.items():
            signal.signal(signal_, original_signal_handler)



from __future__ import annotations

from abc import abstractmethod
from typing import Any, final

from cupyd.core.graph.classes import Node
from cupyd.core.models.node_configuration import FilterConfiguration


class Filter(Node):
    def __init__(self):
        super().__init__()
        self._configuration = FilterConfiguration()

    @abstractmethod
    def filter(self, item: Any) -> Any:
        """Filter an incoming item (Python object)."""

        raise NotImplementedError("Missing implementation of filter() method!")

    @final
    @property
    def configuration(self):
        return self._configuration
